//
//  PlaylistWriter.swift
//  mamba
//
//  Created by David Coufal on 7/12/16.
//  Copyright © 2016 Comcast Cable Communications Management, LLC
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.
//

import Foundation

/// Class responsible for writing in memory HLS playlists to concrete playlists suitable for streaming or other parsers
public class PlaylistWriter {
    
    /**
     Constructs a writer for HLS playlists.
     
     - parameter identityString: This optional string will be used as a HLS 'comment' to easily
     identify playlists parsed and written by this system.
     
     - parameter suppressMambaIdentityString: If set to true, the writer will not add a comment
     identifying that mamba generated the HLS playlist. If set to false (or not set at all) the
     write will output a comment (this is seperate from the identityString above)
     */
    public init(identityString: String? = nil,
                suppressMambaIdentityString: Bool = false) {
        self.identityString = identityString
        self.suppressMambaIdentityString = suppressMambaIdentityString
    }
    
    /// Writes a object implementing `PlaylistInterface` object to a stream. Caller is assumed to be responsible for opening and closing this stream.
    public func write(playlist: PlaylistInterface, toStream stream: OutputStream) throws {
        
        // write initial #EXTM3U
        try write(string: PantosTag.EXTM3U.toString(), toStream: stream)
        
        // write HLS comments to identify ourselves
        if let identityString = identityString {
            try write(string: identityString, toStream: stream)
        }
        if !suppressMambaIdentityString {
            try write(string: " Generated by Mamba(\(Mamba.version)) Copyright (c) 2017 Comcast Corporation", toStream: stream)
        }
        
        // write tags
        for tag in playlist.tags {
            if tag.isDirty {
                guard let writer = playlist.registeredPlaylistTags.writer(forTag: tag.tagDescriptor) else {
                    throw PlaylistWriterError.invalidPlaylist(description: "Cannot write dirty tag with unknown descriptor: \(tag.tagDescriptor.toString())")
                }
                try writer.write(tag: tag, toStream: stream)
            }
            else {
                try GenericTagWriter.write(tag: tag, toStream: stream)
            }
            try stream.write(unicodeScalar: PlaylistTagWritingSeparators.newline)
        }
    }
    
    public func write(playlist: PlaylistInterface) throws -> Data {
        
        let stream = OutputStream.toMemory()
        stream.open()
        
        defer {
            stream.close()
        }
        
        try write(playlist: playlist, toStream: stream)
        if let error = stream.streamError {
            throw OutputStreamError.couldNotWriteToStream(error as NSError)
        }
        guard let data = stream.property(forKey: .dataWrittenToMemoryStreamKey) as? Data else {
            assertionFailure("This method is expected to always return an NSData instance. Update this code to throw a more descriptive error.")
            throw OutputStreamError.couldNotWriteToStream(nil)
        }
        return data
    }

    public func write(playlist: PlaylistInterface) throws -> String {
        let data: Data = try write(playlist: playlist)
        guard let string = String(data: data, encoding: .utf8) else {
            throw OutputStreamError.invalidData(description: "Writing playlist fialure: unable to convert to utf8 coded string")
        }
        return string
    }

    private func write(string: String, toStream stream: OutputStream) throws {
        
        try stream.write(unicodeScalar: PlaylistTagWritingSeparators.hash)
        try stream.write(string: string)
        try stream.write(unicodeScalar: PlaylistTagWritingSeparators.newline)
    }
    
    private let identityString: String?
    private let suppressMambaIdentityString: Bool
}

/// Possible PlaylistWriter error conditions
public enum PlaylistWriterError: Error {
    case invalidPlaylist(description: String?)
}

